__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


import duckdb

from os import path


def get_file_extension(file_path):
    return path.splitext(file_path)[1].lower()


def generate_cottas_info(cottas_file):
    import os
    import datetime

    kv_query = f"SELECT * FROM PARQUET_KV_METADATA('{cottas_file}') WHERE key='index'"
    row_query = f"SELECT num_rows AS triples, num_row_groups AS triples_groups FROM PARQUET_FILE_METADATA('{cottas_file}')"
    properties_query = f"SELECT COUNT(DISTINCT p) FROM PARQUET_SCAN('{cottas_file}')"
    distinct_subjects_query = f"SELECT COUNT(DISTINCT s) FROM PARQUET_SCAN('{cottas_file}')"
    distinct_objects_query = f"SELECT COUNT(DISTINCT o) FROM PARQUET_SCAN('{cottas_file}')"
    schema_query = f"DESCRIBE SELECT * FROM PARQUET_SCAN('{cottas_file}') LIMIT 1"
    compression_query = f"SELECT compression FROM PARQUET_METADATA('{cottas_file}')"

    (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime) = os.stat(cottas_file)
    cottas_issued = datetime.datetime.fromtimestamp(ctime).isoformat()

    info_dict = {
        'index': duckdb.execute(kv_query).fetchone()[2].decode(),
        'triples': duckdb.execute(row_query).fetchone()[0],
        'triples_groups': duckdb.execute(row_query).fetchone()[1],
        'properties': duckdb.execute(properties_query).fetchone()[0],
        'distinct_subjects': duckdb.execute(distinct_subjects_query).fetchone()[0],
        'distinct_objects': duckdb.execute(distinct_objects_query).fetchone()[0],
        'issued': cottas_issued,
        'size (MB)': os.path.getsize(cottas_file) / 10**6,
        'compression': duckdb.execute(compression_query).fetchone()[0],
        'quads': 'g' in [res[0] for res in duckdb.execute(schema_query).fetchall()],
    }

    return info_dict


def is_valid_index(index):
    index = index.lower()
    if len(index) == 3:
        if set(index) != {'s', 'p', 'o'}:
            return False
    elif len(index) == 4:
        if set(index) != {'s', 'p', 'o', 'g'}:
            return False
    else:
        return False
    return True


def verify_cottas_file(cottas_file):
    verify_query = f"DESCRIBE SELECT * FROM PARQUET_SCAN('{cottas_file}') LIMIT 1"

    cottas_columns = set()
    for res in duckdb.execute(verify_query).fetchall():
        cottas_columns.add(res[0])

    for pos in ['s', 'p', 'o']:
        if pos not in cottas_columns:
            return False

    return cottas_columns <= {'s', 'p', 'o', 'g'}



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


import duckdb
import pyoxigraph

from rdflib.util import from_n3
from .tp_translator import translate_triple_pattern
from .utils import verify_cottas_file


class COTTASDocument():
    def __init__(self, path: str):
        if not verify_cottas_file(path):
            raise Exception(f"{path} is not a valid COTTAS file.")

        self._cottas_path = path
        duckdb.query(
            f"SET parquet_metadata_cache=true; SET enable_progress_bar=false; SELECT * FROM PARQUET_SCAN('{path}')")
        self._num_triples = duckdb.execute(f"SELECT COUNT(*) FROM PARQUET_SCAN('{path}')").fetchone()[0]
        self._is_quad_table = 'g' in [res[0] for res in duckdb.execute(
            f"DESCRIBE SELECT * FROM PARQUET_SCAN('{path}') LIMIT 1").fetchall()]

    @property
    def cottas_file(self) -> str:
        """The COTTAS file path."""
        return self._cottas_path

    @property
    def is_quad_table(self) -> str:
        """Whether the table has named graphs."""
        return self._is_quad_table

    def __len__(self, context) -> int:
        """The number of RDF triples in the COTTAS store."""
        return self._num_triples

    @property
    def nb_subjects(self) -> int:
        """The number of subjects in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT s) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    @property
    def nb_predicates(self) -> int:
        """The number of predicates in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT p) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    @property
    def nb_objects(self) -> int:
        """The number of objects in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT o) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    def search(self, pattern, limit=None, offset=None, results_in_n3=True):
        if len(pattern) == 4 and not self._is_quad_table:
            raise Exception("The COTTAS file is not a quad table, quad patterns are not valid.")

        triples = duckdb.execute(translate_triple_pattern(self._cottas_path, pattern, limit, offset)).fetchall()

        if results_in_n3:
            return triples

        for i, triple in enumerate(triples):
            if len(triple) == 3:
                triples[i] = (from_n3(triple[0]), from_n3(triple[1]), from_n3(triple[2]))
            else:
                # len(triple) = 4
                triples[i] = (from_n3(triple[0]), from_n3(triple[1]), from_n3(triple[2]), from_n3(triple[3]))

        return triples




__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


from.constants import i_pos


def _parse_tp(tp_str):
    #statement = "Select ?s ?p ?o where {?s ?p ?o.}"
    #query_tree = parser.parseQuery(statement)
    #q_algebra = algebra.translateQuery(query_tree)
    #algebra.pprintAlgebra(q_algebra)

    splitted_tp_str = tp_str.split()
    s_term = splitted_tp_str[0]
    p_term = splitted_tp_str[1]
    o_term = tp_str.replace(s_term, '', 1).replace(p_term, '', 1).strip()
    g_term = None

    # check whether it is a quad pattern
    if ' <' in o_term:
        splitted_o_term = o_term.split()
        g_term = splitted_o_term[len(splitted_o_term)-1].strip()
        o_term = o_term.replace(g_term, '').strip()

    if g_term:
        return [s_term, p_term, o_term, g_term]
    else:
        return [s_term, p_term, o_term]


def translate_triple_pattern(cottas_file, tp, limit=None, offset=None):
    """
    Given a COTTAS file and a user-defined triple pattern, translate the triple pattern to an SQL query over COTTAS.

    :param cottas_file: path to a COTTAS file
    :param tp: a user-defined triple pattern
    :return: SQL query for the triple pattern
    """

    if type(tp) is str:
        tp_tuple = _parse_tp(tp)
    else:
        tp_aux = []
        for i in range(len(tp)):
            if tp[i] is None:
                tp_aux.append(f"?{i_pos[i]}")
            else:
                tp_aux.append(tp[i].n3())
        tp_tuple = tp_aux

    if len(tp_tuple) == 3:
        tp_query = f"SELECT s, p, o FROM PARQUET_SCAN('{cottas_file}') WHERE "
    elif len(tp_tuple) == 4:
        tp_query = f"SELECT s, p, o, g FROM PARQUET_SCAN('{cottas_file}') WHERE "
    else:
        raise TypeError("The pattern must be a tuple of length 3 (triple) or 4 (quad).")

    # build selection iterating over all positions in the triple pattern
    for i in range(4):
        # skip named graph if not in the triple pattern
        if i < len(tp_tuple):
            if not tp_tuple[i].startswith('?'):
                # scape 'quotes'
                tp_tuple[i] = tp_tuple[i].replace("'", "''")
                tp_query += f"{i_pos[i]}='{tp_tuple[i]}' AND "

    # remove final `AND ` and `WHERE `
    if tp_query.endswith('AND '):
        tp_query = tp_query[:-4]
    if tp_query.endswith('WHERE '):
        tp_query = tp_query[:-6]

    # handle limit and offset
    if limit:
        if type(limit) is not int:
            raise TypeError("Limit must be an integer.")
        tp_query += f" LIMIT {limit}"
    if offset:
        if type(offset) is not int:
            raise TypeError("Offset must be an integer.")
        tp_query += f" OFFSET {offset}"

    return tp_query


def translate_triple_pattern_tuple(cottas_file, tp_tuple, limit=None, offset=None):
    """
    Given a COTTAS file and an RDFlib triple pattern tuple, translate the triple pattern to an SQL query over COTTAS.

    :param cottas_file: path to a COTTAS file
    :param tp_str: a user-defined triple pattern
    :return: SQL query for the triple pattern
    """

    if len(tp_tuple) == 3:
        tp_query = f"SELECT s, p, o FROM PARQUET_SCAN('{cottas_file}') WHERE "
    elif len(tp_tuple) == 4:
        tp_query = f"SELECT s, p, o, g FROM PARQUET_SCAN('{cottas_file}') WHERE "
    else:
        raise TypeError("The pattern must be a tuple of length 3 (triple) or 4 (quad).")

    # build selection iterating over all positions in the triple pattern
    for i in range(4):
        # skip named graph if not in the triple pattern
        if i < len(tp_tuple):
            if not tp_tuple[i] is None:
                if type(tp_tuple[i]) is str:
                    tp_query += f"{i_pos[i]}='{tp_tuple[i]}' AND "
                else:
                    tp_query += f"{i_pos[i]}='{tp_tuple[i].n3()}' AND "

    # remove final `AND ` and `WHERE `
    if tp_query.endswith('AND '):
        tp_query = tp_query[:-4]
    if tp_query.endswith('WHERE '):
        tp_query = tp_query[:-6]

    # handle limit and offset
    if limit:
        if type(limit) is not int:
            raise TypeError("Limit must be an integer.")
        tp_query += f" LIMIT {limit}"
    if offset:
        if type(offset) is not int:
            raise TypeError("Offset must be an integer.")
        tp_query += f" OFFSET {offset}"

    return tp_query



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


import duckdb
import pyoxigraph
import os
import pandas as pd

from random import randint

from .constants import file_ext_2_mime_type
from .tp_translator import translate_triple_pattern
from .utils import generate_cottas_info, get_file_extension, is_valid_index, verify_cottas_file
from .cottas_store import COTTASStore
from .cottas_document import COTTASDocument


def rdf2cottas(rdf_file_path, cottas_file_path, index='spo', disk=False):
    if index and not is_valid_index(index):
        print(f"Index `{index}` is not valid.")
        return

    mime_type = file_ext_2_mime_type[get_file_extension(file_path=rdf_file_path)]
    quad_found = False

    db_connection = 'pycottas.duckdb' if disk else ':memory:'
    triplestore = duckdb.connect(db_connection)

    create_query = """
                CREATE TABLE quads (s VARCHAR NOT NULL, p VARCHAR NOT NULL, o VARCHAR NOT NULL, g VARCHAR)
            """
    triplestore.execute(create_query)
    triplestore.query("SET preserve_insertion_order = false; SET enable_progress_bar=false;")

    quads = []
    i = 0
    for quad in pyoxigraph.parse(rdf_file_path, base_iri=None, mime_type=mime_type):
        quad = [str(term) for term in quad]

        if len(quad) == 3:
            # for empty quad
            quad.append(None)
        else:
            quad_found = True
        quads.append(quad)

        if i == 1000000:
            # bulk add quads
            # bulk add quads
            quads_df = pd.DataFrame.from_records(quads, columns=['st', 'pt', 'ot', 'gt'])
            temporal_table = f'temporal_quads_{randint(0, 1000000)}'
            triplestore.register(temporal_table, quads_df)
            insert_query = f"INSERT INTO quads (SELECT st, pt, ot, gt FROM {temporal_table})"
            triplestore.execute(insert_query)
            triplestore.unregister(temporal_table)

            # reset quads
            quads = []
            i = 0
        else:
            i += 1

    # bulk add quads
    quads_df = pd.DataFrame.from_records(quads, columns=['st', 'pt', 'ot', 'gt'])
    temporal_table = f'temporal_quads_{randint(0, 1000000)}'
    triplestore.register(temporal_table, quads_df)
    insert_query = f"SET preserve_insertion_order = false; INSERT INTO quads (SELECT st, pt, ot, gt FROM {temporal_table})"
    triplestore.execute(insert_query)
    triplestore.unregister(temporal_table)

    # export the triple table
    if quad_found:
        export_query = f"COPY (SELECT DISTINCT s, p, o, g FROM quads"
    else:
        export_query = f"COPY (SELECT DISTINCT s, p, o FROM quads"
    if index:
        export_query += " ORDER BY "
        for p in index:
            export_query += f"{p}, "
        export_query = export_query[:-2]
    export_query += f") TO '{cottas_file_path}' (FORMAT PARQUET, COMPRESSION ZSTD, COMPRESSION_LEVEL 22, PARQUET_VERSION v2, KV_METADATA {{index: '{index.lower()}'}})"
    triplestore.execute(export_query)


def cottas2rdf(cottas_file_path, rdf_file_path):
    f = open(rdf_file_path, 'w')
    duckdb.query("SET enable_progress_bar=false;")

    has_named_graph = 'g' in list(duckdb.query(f"SELECT name FROM PARQUET_SCHEMA('{cottas_file_path}')").df()['name'])
    cur = duckdb.execute(f"SELECT s, p, o{', g' if has_named_graph else ''} FROM PARQUET_SCAN('{cottas_file_path}')")
    cur_chunk_df = cur.fetch_df_chunk()
    while len(cur_chunk_df):
        quads = cur_chunk_df.values.tolist()
        for quad in quads:
            if has_named_graph:
                # in case of named graph
                quad = f"{quad[0]} {quad[1]} {quad[2]} {quad[3]}"
            else:
                quad = f"{quad[0]} {quad[1]} {quad[2]}"
            f.write(f'{quad} .\n')

        cur_chunk_df = cur.fetch_df_chunk()
    f.close()


def search(cottas_file_path, triple_pattern):
    return duckdb.execute(translate_triple_pattern(f"{cottas_file_path}", triple_pattern)).fetchall()


def cat(cottas_file_paths, cottas_cat_file_path, index='spo', remove_input_files=False):
    if index and not is_valid_index(index):
        print(f"Index `{index}` is not valid.")
        return

    cat_query = f"COPY (SELECT DISTINCT s, p, o FROM PARQUET_SCAN({cottas_file_paths}, union_by_name = true)"
    if index:
        cat_query += " ORDER BY "
        for p in index:
            cat_query += f"{p}, "
        cat_query = cat_query[:-2]
    cat_query += f") TO '{cottas_cat_file_path}' (FORMAT PARQUET, COMPRESSION ZSTD, COMPRESSION_LEVEL 22, PARQUET_VERSION v2, KV_METADATA {{index: '{index.lower()}'}})"
    duckdb.execute(cat_query)

    if remove_input_files:
        for file in cottas_file_paths:
            os.remove(file)


def diff(cottas_file_1_path, cottas_file_2_path, cottas_diff_file_path, index='spo', remove_input_files=False):
    if index and not is_valid_index(index):
        print(f"Index `{index}` is not valid.")
        return

    diff_query = f"COPY (SELECT * FROM (SELECT DISTINCT * FROM PARQUET_SCAN('{cottas_file_1_path}') EXCEPT SELECT * FROM PARQUET_SCAN('{cottas_file_2_path}'))"
    if index:
        diff_query += " ORDER BY "
        for p in index:
            diff_query += f"{p}, "
        diff_query = diff_query[:-2]
    diff_query += f") TO '{cottas_diff_file_path}' (FORMAT PARQUET, COMPRESSION ZSTD, COMPRESSION_LEVEL 22, PARQUET_VERSION v2, KV_METADATA {{index: '{index.lower()}'}})"
    duckdb.execute(diff_query)

    if remove_input_files:
        os.remove(cottas_file_1_path)
        os.remove(cottas_file_2_path)


def info(cottas_file_path):
    return generate_cottas_info(cottas_file_path)


def verify(cottas_file_path):
    return verify_cottas_file(cottas_file_path)



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


from pandas import DataFrame
from argparse import ArgumentParser

from .__init__ import *
from.constants import i_pos


EPILOG_TEXT = 'Copyright © 2023 Julián Arenas-Guerrero'


if __name__ == "__main__":
    parser = ArgumentParser(
        prog='pycottas',
        epilog=EPILOG_TEXT)

    subparsers = parser.add_subparsers(help='subcommand help', dest='subparser_name', required=True)

    parse_rdf2cottas = subparsers.add_parser('rdf2cottas', help='Compress an RDF file into COTTAS format', epilog=EPILOG_TEXT)
    parse_rdf2cottas.add_argument('-r', '--rdf_file', type=str, required=True, help='Path to RDF file')
    parse_rdf2cottas.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to COTTAS file')
    parse_rdf2cottas.add_argument('-i', '--index', type=str, required=False, default='SPO', help='Zonemap index, e.g.: `SPO`, `PSO`, `GPOS`')
    parse_rdf2cottas.add_argument('-d', '--disk', type=bool, required=False, default=False, help='Whether to use on-disk storage')

    parse_cottas2rdf = subparsers.add_parser('cottas2rdf', help='Decompress a COTTAS file to RDF (N-Triples)', epilog=EPILOG_TEXT)
    parse_cottas2rdf.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to COTTAS file')
    parse_cottas2rdf.add_argument('-r', '--rdf_file', type=str, required=True, help='Path to RDF file (N-Triples)')

    parse_cottas2rdf = subparsers.add_parser('search', help='Evaluate a triple pattern', epilog=EPILOG_TEXT)
    parse_cottas2rdf.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to COTTAS file')
    parse_cottas2rdf.add_argument('-t', '--triple_pattern', type=str, required=True, help='Triple pattern, e.g., `?s <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> ?o`')
    parse_cottas2rdf.add_argument('-r', '--result_option', default='tuples', choices=['tuples', 'table', 'to_csv'], help='What to do with the result set')

    parse_info = subparsers.add_parser('info', help='Get the metadata of a COTTAS file', epilog=EPILOG_TEXT)
    parse_info.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to COTTAS file')

    parse_verify = subparsers.add_parser('verify', help='Check whether a file is a valid COTTAS file', epilog=EPILOG_TEXT)
    parse_verify.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to COTTAS file')

    parse_cat = subparsers.add_parser('cat', help='Merge multiple COTTAS files', epilog=EPILOG_TEXT)
    parse_cat.add_argument('--input_cottas_files', type=str, nargs='+', required=True, help='Path of the input COTTAS files')
    parse_cat.add_argument('--output_cottas_file', type=str, required=True, help='Path of the output COTTAS file')
    parse_cat.add_argument('-i', '--index', type=str, required=False, default='SPO', help='Zonemap index, e.g.: `SPO`, `PSO`, `GPOS`')
    parse_cat.add_argument('-r', '--remove_input_files', type=bool, required=False, default=False, help='Whether to remove input COTTAS files after merging')

    parse_diff = subparsers.add_parser('diff', help='Subtract the triples in a COTTAS files from another', epilog=EPILOG_TEXT)
    parse_diff.add_argument('-c', '--cottas_file', type=str, required=True, help='Path to the COTTAS file')
    parse_diff.add_argument('-s', '--subtract_cottas_file', type=str, required=True, help='Path to the COTTAS file to subtract')
    parse_diff.add_argument('-o', '--output_cottas_file', type=str, required=True, help='Path to the output COTTAS file')
    parse_diff.add_argument('-i', '--index', type=str, required=False, default='SPO', help='Zonemap index, e.g.: `SPO`, `PSO`, `GPOS`')
    parse_diff.add_argument('-r', '--remove_input_files', type=bool, required=False, default=False, help='Whether to remove the input COTTAS files after merging')

    args = parser.parse_args()

    if args.subparser_name == 'rdf2cottas':
        rdf2cottas(args.rdf_file, args.cottas_file, args.index, args.disk)

    elif args.subparser_name == 'cottas2rdf':
        cottas2rdf(args.cottas_file, args.rdf_file)

    elif args.subparser_name == 'search':
        if args.result_option == 'table':
            print(duckdb.query(translate_triple_pattern(f"{args.cottas_file}", args.triple_pattern)))
        elif args.result_option == 'to_csv':
            duckdb.query(translate_triple_pattern(f"{args.cottas_file}", args.triple_pattern)).df().to_csv(
                'cottas_search.csv', index=False)
        elif args.result_option == 'tuples':
            print(search(args.cottas_file, args.triple_pattern))

    elif args.subparser_name == 'info':
        print(info(args.cottas_file))

    elif args.subparser_name == 'verify':
        print(verify(args.cottas_file))

    elif args.subparser_name == 'cat':
        cat(args.input_cottas_files, args.output_cottas_file, args.index, args.remove_input_files)

    elif args.subparser_name == 'diff':
        index = args.index if args.index else 'spo'
        remove_input_files = args.remove_input_files if args.remove_input_files else False
        diff(args.cottas_file, args.subtract_cottas_file, args.output_cottas_file, args.index, args.remove_input_files)



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


import duckdb

from typing import Iterable
from rdflib.store import Store
from rdflib.util import from_n3

from .utils import verify_cottas_file
from .types import Triple
from .tp_translator import translate_triple_pattern_tuple


class COTTASStore(Store):
    """An implementation of a Store over a COTTAS document.

    It is heavily inspired by the work from @FlorianLudwig (https://github.com/RDFLib/rdflib/issues/894) and adapted
    from rdflib-hdt (https://github.com/RDFLib/rdflib-hdt).

    Args:
      - path: Absolute path to the COTTAS file to load.
    """
    def __init__(self, path: str, configuration=None, identifier=None):
        super(COTTASStore, self).__init__(configuration=configuration, identifier=identifier)

        if not verify_cottas_file(path):
            raise Exception(f"{path} is not a valid COTTAS file.")

        self._cottas_path = path
        duckdb.query(
            f"SET parquet_metadata_cache=true; SET enable_progress_bar=false; SELECT * FROM PARQUET_SCAN('{path}')")
        self._num_triples = duckdb.execute(f"SELECT COUNT(*) FROM PARQUET_SCAN('{path}')").fetchone()[0]
        self._is_quad_table = 'g' in [res[0] for res in duckdb.execute(
            f"DESCRIBE SELECT * FROM PARQUET_SCAN('{path}') LIMIT 1").fetchall()]

    @property
    def cottas_file(self) -> str:
        """The COTTAS file path."""
        return self._cottas_path

    @property
    def is_quad_table(self) -> str:
        """Whether the table has named graphs."""
        return self._is_quad_table

    def __len__(self, context) -> int:
        """The number of RDF triples in the COTTAS store."""
        return self._num_triples

    @property
    def nb_subjects(self) -> int:
        """The number of subjects in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT s) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    @property
    def nb_predicates(self) -> int:
        """The number of predicates in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT p) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    @property
    def nb_objects(self) -> int:
        """The number of objects in the COTTAS store."""
        return duckdb.execute(f"SELECT COUNT(DISTINCT o) FROM PARQUET_SCAN('{self._cottas_path}')").fetchone()[0]

    def triples(self, pattern, context) -> Iterable[Triple]:
        """Search for a triple pattern in a COTTAS store.

        Args:
          - pattern: The triple pattern (s, p, o) to search.
          - context: The query execution context.

        Returns: An iterator that produces RDF triples matching the input triple pattern.
        """
        for triple in duckdb.execute(translate_triple_pattern_tuple(self._cottas_path, pattern)).fetchall():
            triple = from_n3(triple[0]), from_n3(triple[1]), from_n3(triple[2])
            yield triple, None
        return

    def create(self, configuration):
        raise TypeError('The COTTAS store is read only!')

    def destroy(self, configuration):
        raise TypeError('The COTTAS store is read only!')

    def commit(self):
        raise TypeError('The COTTAS store is read only!')

    def rollback(self):
        raise TypeError('The COTTAS store is read only!')

    def add(self, _, context=None, quoted=False):
        raise TypeError('The COTTAS store is read only!')

    def addN(self, quads):
        raise TypeError('The COTTAS store is read only!')

    def remove(self, _, context):
        raise TypeError('The COTTAS store is read only!')



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "julian.arenas.guerrero@upm.es"


from typing import Optional, Set, Tuple, Union
from rdflib import Literal, URIRef, Variable

Term = Union[URIRef, Literal]
Triple = Tuple[Term, Term, Term]
TriplePattern = Union[URIRef, Literal, Variable]
SearchQuery = Tuple[Optional[Term], Optional[Term], Optional[Term]]
BGP = Set[TriplePattern]